extends RigidBody2D

const Colors = preload("res://Constants/Colors.gd")
const DamageDealer = preload("res://Actors/DamageDealer.gd")
const Ping = preload("res://Items/Ping/Ping.tscn")
const PopText = preload("res://Juice/Text/PopText.tscn")
const Trigonometry = preload("res://Util/Trigonometry.gd")

export(GDScript) var starting_state
export(String) var name
export(String) var description
export(int) var health
export(int) var xp
export(int, FLAGS, "Blunt", "Slash", "Pierce", "Fire", "Ice", "Wind", "Earth", "Light", "Dark") var resistances = 0
export(int, FLAGS, "Blunt", "Slash", "Pierce", "Fire", "Ice", "Wind", "Earth", "Light", "Dark") var weaknesses = 0
export(StringArray) var drop_paths
export(IntArray) var drop_rates
export(bool) var process = false
export(bool) var fixed_process = false

var hp = null
var state = null
var temporarily_ignore_collisions = []

onready var _hitbox = get_node("Hitbox")

# Initialization. This is called automatically before the child class initialization is called.
func _ready():
	# Make sure we didn't forget any important properties in the editor.
	assert(name != null)
	assert(description != null)
	assert(health != null)
	assert(xp != null)
	assert((drop_paths == null and drop_rates == null) or drop_paths.size() == drop_rates.size())
	hp = health
	if starting_state != null:
		state = starting_state.new()
		state.enter(self)
	_hitbox.add_to_group("DamagesPlayer")
	_hitbox.add_to_group("Enemy")
	_hitbox.add_to_group("Grabbable")
	_hitbox.connect("area_enter", self, "_on_area_enter")
	if process == true:
		set_process(true)
	if fixed_process == true:
		set_fixed_process(true)

func _process(delta):
	if state != null:
		state.state_loop("_process", delta)

func _fixed_process(delta):
	if state != null:
		state.state_loop("_fixed_process", delta)

# Receive damage when appropriate.
func _on_area_enter(area):
	if not area.is_in_group("DamagesEnemies"):
		return
	take_damage(area)

# Take damage and physics effects, factoring in types.
func take_damage(damage_dealer):
	var damage = damage_dealer.get_damage()
	# 50% increase or decrease per matched type.
	var dealt_types = damage_dealer.get_types()
	var net_effectiveness = 0
	for type in DamageDealer.TYPES:
		if type & resistances and type & dealt_types:
			damage = damage / DamageDealer.EFFECTIVENESS
			net_effectiveness = net_effectiveness - 1
		if type & weaknesses and type & dealt_types:
			damage = damage * DamageDealer.EFFECTIVENESS
			net_effectiveness = net_effectiveness + 1
	damage = ceil(damage)
	hp = hp - damage
	# Visually represent the damage.
	var poptext = PopText.instance()
	var label = poptext.get_node("Label")
	label.set_text(str(damage))
	if net_effectiveness < 0:
		label.set("custom_colors/font_color", Colors.ENEMY_DAMAGE_RESIST)
	elif net_effectiveness > 0:
		label.set("custom_colors/font_color", Colors.ENEMY_DAMAGE_WEAK)
	else:
		label.set("custom_colors/font_color", Colors.ENEMY_DAMAGE_NORMAL)
	poptext.set_global_pos(Trigonometry.get_global_point_of_overlap(damage_dealer, _hitbox))
	Global.get_level().add_child(poptext)
	# Go flying style logic.
	if damage > health:
		var impact = damage_dealer.get_impact() * ((5.0 * damage) / health)
		if impact > 0:
			var origin = damage_dealer.player.get_global_pos()
			var pos = get_global_pos()
			var delta_vector = origin - pos
			var launch_angle = atan2(-delta_vector.y, -delta_vector.x)
			apply_impulse(Vector2(0, 0), Vector2(impact * cos(launch_angle), -impact * sin(launch_angle)))
	if hp <= 0:
		die()

# Expire.
func die():
	add_collision_exception_with(Global.get_player())
	# No longer take damage.
	get_node("Hitbox").disconnect("area_enter", self, "_on_area_enter")
	# Update the Bestiary.
	var path = get_filename()
	var entry = Bestiary.data[path]
	entry.killed = entry.killed + 1
	# Potentially spawn items.
	for drop_chance in entry.drops:
		var drop = drop_chance.roll()
		if drop != null:
			var level = Global.get_level()
			var pos = get_global_pos()
			drop.set_global_pos(pos)
			level.add_child(drop)
			var ping = Ping.instance()
			ping.set_global_pos(pos)
			level.add_child(ping)
	# TODO: Optional dying animation / effects.
	queue_free()

# Call the given function of the given state.
func invoke(state, function):
	assert(self.state.get_name() == state)
	self.state.call(function)
