extends "res://Actors/Enemies/EnemyState.gd"

const Bone = preload("res://Actors/Enemies/Graveyard/Skeleton/Bone.tscn")
const Trigonometry = preload("res://Util/Trigonometry.gd")

const SPEED = 100

var _face_right = true
var _move_right = true

var _animation_player = null
var _back_checker = null
var _front_checker = null
var _sprites = null

var _throw_bone = false
var _bone_timer = null
var _jump_timer = null
var _stop_timer = null
var _turnaround_timer = null

func get_name():
	return "Frolicking"

func enter(parent):
	.enter(parent)
	_animation_player = parent.get_node("AnimationPlayer")
	_back_checker = parent.get_node("BackChecker")
	_front_checker = parent.get_node("FrontChecker")
	_sprites = parent.get_node("Sprites")
	parent.get_node("Checkbox").connect("area_enter", self, "_on_area_enter")
	parent.get_node("Checkbox").connect("area_exit", self, "_on_area_exit")
	_reset_bone_timer()
	_reset_jump_timer()
	_reset_stop_timer()
	_reset_turnaround_timer()

# Throw bones at the player every so often when they're in our area.
func _on_area_enter(area):
	if area.is_in_group("Player"):
		_throw_bone = true

func _on_area_exit(area):
	if area.is_in_group("Player"):
		_throw_bone = false

func _process(delta):
	# Manage timers.
	_bone_timer = _bone_timer - delta
	_jump_timer = _jump_timer - delta
	_stop_timer = _stop_timer - delta
	# Stop periodically.
	if _stop_timer <= 0:
		_reset_stop_timer()
		x_velocity = 0
		_animation_player.play("Standing")
	_turnaround_timer = _turnaround_timer - delta
	# Turn around periodically.
	if _turnaround_timer <= 0:
		_reset_turnaround_timer()
		if parent.get_pos().y == initial_position.y:
			_move_right = not _move_right
			if _move_right == true:
				x_velocity = SPEED
			else:
				x_velocity = -SPEED
			if _animation_player.get_current_animation() != "Walking":
				_animation_player.play("Walking")
	# Update facing.
	if Global.get_player().get_global_pos().x < parent.get_global_pos().x:
		if _face_right == true:
			_face_right = false
			_sprites.set_scale(_sprites.get_scale() * Vector2(-1, 1))
	else:
		if _face_right == false:
			_face_right = true
			_sprites.set_scale(_sprites.get_scale() * Vector2(-1, 1))

func _fixed_process(delta):
	# Throw bones?
	if _throw_bone and _bone_timer <= 0:
		_throw_a_bone()
	# Jump / fall.
	_handle_jumps(delta)
	# Wander.
	_wander(delta)
	._fixed_process(delta)

# Kick off the bone-throwing animation.
func _throw_a_bone():
	_reset_bone_timer()
	x_velocity = 0
	_reset_stop_timer()
	_reset_jump_timer()
	_reset_turnaround_timer()
	_animation_player.play("Throwing")

# Spawn the bone projectile (called from AnimationPlayer).
func throw():
	var bone = Bone.instance()
	var bone_pos = parent.get_node("Sprites/Hips/Torso/FrontShoulder/FrontArm/FrontHand").get_global_pos()
	bone.set_global_pos(bone_pos)
	# Negative angles because of inverted Y axis.
	var angle = - 1.39626 # 80 deg.
	if _face_right == false:
		angle = - 1.91986 # 110 deg.
	var velocity = Trigonometry.get_velocity_of_arc_intercept(bone, Global.get_player().get_global_pos(), angle)
	bone.set_linear_velocity(velocity)
	Global.get_level().add_child(bone)

# Actually jump. Should be invoked by AnimationPlayer.
func _jump():
	x_velocity = 0
	y_velocity = -SPEED * 2

# Trigger jumps and handle falling.
func _handle_jumps(delta):
	var parent_pos = parent.get_pos()
	# Reset timer.
	if _jump_timer <= 0:
		_reset_jump_timer()
		# Play the jumping animation if we're grounded.
		if parent_pos.y == initial_position.y:
			_animation_player.play("Jumping")
			x_velocity = 0
			# Don't let other timers interrupt us.
			_turnaround_timer = _turnaround_timer + 3
			_stop_timer = _stop_timer + 3
	# Don't sink lower than our initial position.
	if parent_pos.y > initial_position.y:
		y_velocity = 0
		parent.set_pos(Vector2(parent_pos.x, initial_position.y))
		# And restart their movement.
		if _move_right == true:
			x_velocity = SPEED
		else:
			x_velocity = -SPEED
		_animation_player.play("Walking")
	# Otherwise fall.
	elif parent_pos.y < initial_position.y:
		y_velocity = y_velocity + (SPEED * 2 * delta)

func _wander(delta):
	# Don't wander if we're not on level with our starting position.
	if parent.get_pos().y != initial_position.y:
		return
	# Update facing.
	if _move_right == true:
		if not _front_checker.is_overlapping():
			if _back_checker.is_overlapping():
				x_velocity = -SPEED
				_reset_turnaround_timer()
				_move_right = false
			else:
				return # Airborne.
	else:
		if not _back_checker.is_overlapping():
			if _front_checker.is_overlapping():
				x_velocity = SPEED
				_reset_turnaround_timer()
				_move_right = true

# Reset the animation, because we shamefully did not use a state machine.
func stand():
	_animation_player.play("Standing")

func _reset_bone_timer():
	_bone_timer = 3 + (randf() * 5)

func _reset_jump_timer():
	_jump_timer = (randi() % 10) + 5

func _reset_stop_timer():
	_stop_timer = (randi() % 8) + 2

func _reset_turnaround_timer():
	_turnaround_timer = (randi() % 5) + 2

func exit():
	# Tidy up our connection.
	parent.get_node("Checkbox").disconnect("area_enter", self, "_on_area_enter")
