extends KinematicBody2D

# Classes to instantiate.
const Colors = preload("res://Constants/Colors.gd")
const DamageDealer = preload("res://Actors/DamageDealer.gd")
const Map = preload("res://UI/Map/Map.tscn")
const Menu = preload("res://UI/Menu/Menu.tscn")
const PlatformMoving = preload("res://Actors/Player/Substates/PlatformMoving.gd")
const PopText = preload("res://Juice/Text/PopText.tscn")
const Standing = preload("res://Actors/Player/States/Standing.gd")
const Trigonometry = preload("res://Util/Trigonometry.gd")

# Constants
const DEFAULT_HORIZONTAL_SPEED = 180.0 # pixels/s
const DEFAULT_JUMP_HEIGHT = -150 # Pixels
const DEFAULT_RUNNING_MULTIPLIER = 2.0
const DUCKING_DIVISOR = 1.75
const JOYSTICK_THRESHOLD = 0.2
const MAX_JUMP_HEIGHT_MULTIPLIER = 3.0
const VERTICAL_GRACE_FACTOR = 20.0
const ZERO_VECTOR = Vector2(0, 0)

# Stats. Default 100.
var strength = 100 # Sends enemies flying.
var constitution = 100 # Lets you trample low-enough level enemies beneath your feet.
var agility = 100 setget _set_agi # Makes everything you do faster.
var mind = 100 # Scales up the size of weapons.

# Object reference variables. Don't onready since we're shared between scenes.
var attack_animation_player = null
var animation_player = null
var state = null
var substates = []
var weapon = null
var _ceiling_checker = null
var _floor_checker = null
var _grabsphere = null
var _hitbox = null
var _slope_checker_L = null
var _slope_checker_R = null
var _sprite = null
var _wall_checker_L = null
var _wall_checker_R = null
var _state_node = null

# Public variables.

var attack_primary = "Punch"
var attack_secondary = "Punch"
var can_dodge = true # The ability to dodge in general.
var can_double_jump = true # The ability to double-jump in general.
var can_glide = true
var can_swim = false
var direction = 1 setget set_direction # 1 for Right, -1 for Left.
var has_dodge = false # The ability to dodge right now.
var has_extra_jump = false # The capacity to double-jump right now.
var horizontal_speed = DEFAULT_HORIZONTAL_SPEED
var jump_height = DEFAULT_JUMP_HEIGHT
var jump_height_multiplier = 1.0
var last_movement_vector = Vector2(0, 0)
var running_multiplier = DEFAULT_RUNNING_MULTIPLIER
var temporarily_ignore_collisions = []
var walking_multiplier = 1.0

# Private variables.
var _attack_is_flipped = {}
var _could_alternate_attack = false
var _horizontal_multiplier = 1.0
var _last_attack_front_hand = true
var _raycast_exceptions = null
var _vertical_drop_threshold = null

# Shape sizes.
var _collision_shape_extents = null
var _hitbox_extents = null
var _wall_checker_extents = null # Assumes symmetry.

# Collect references and initialize state.
func _ready():
	# If we're passed from a previous level, we don't need to re-initialize.
	if state != null:
		return
	set_fixed_process(true)
	set_process(true)
	set_process_input(true)
	load("res://Actors/Player/Substates/Fighting.gd").new().attach(self)
	attack_animation_player = get_node("AttackAnimationPlayer")
	attack_animation_player.connect("finished", self, "_attack_animation_finished")
	# Populate the flipped animation dictionary with each attack animation name.
	for animation in attack_animation_player.get_animation_list():
		_attack_is_flipped[animation] = false
	animation_player = get_node("AnimationPlayer")
	_ceiling_checker = get_node("CeilingChecker")
	_floor_checker = get_node("FloorChecker")
	_grabsphere = get_node("GrabSphere")
	_hitbox = get_node("Hitbox")
	_hitbox.add_to_group("CollectsItems")
	_hitbox.add_to_group("Player")
	_hitbox.connect("area_enter", self, "_hitbox_area_enter")
	_slope_checker_L = get_node("SlopeCheckerL")
	_slope_checker_R = get_node("SlopeCheckerR")
	_sprite = get_node("Sprite")
	_wall_checker_L = get_node("WallCheckerL")
	_wall_checker_R = get_node("WallCheckerR")
	_state_node = get_node("State")
	# Stuff to help determine if there's ground just below us.
	_vertical_drop_threshold = get_node("CollisionShape2D").get_shape().get_extents().y
	_raycast_exceptions = []
	_raycast_exceptions.append(self)
	for child in get_children():
		if child extends CollisionObject2D:
			_raycast_exceptions.append(child)
	# Get starting shape sizes.
	_collision_shape_extents = get_node("CollisionShape2D").get_shape().get_extents()
	_hitbox_extents = get_node("Hitbox/CollisionShape2D").get_shape().get_extents()
	_wall_checker_extents = get_node("WallCheckerL/CollisionShape2D").get_shape().get_extents()
	# Initialize state.
	state = Standing.new()
	state.enter(self)

# The current state gets to intercept input events.
func _input(event):
	for substate in substates:
		substate._input(event)
	if (event.type == InputEvent.KEY or event.type == InputEvent.JOYSTICK_BUTTON) and event.is_pressed() and not event.is_echo():
		# Quit on Escape.
		if event.is_action("ui_cancel"):
			get_tree().quit()
		# Bring up a map.
		if event.is_action("ui_focus_next"):
			get_tree().set_input_as_handled()
			Global.get_level().get_node("HUD").add_child(Map.instance())
			return
		# Bring up a menu.
		if event.is_action("ui_accept"):
			get_tree().set_input_as_handled()
			Global.get_level().get_node("HUD").add_child(Menu.instance())
			return
		# Only allow attacks during certain states.
		if can_attack():
			if event.is_action("player_attack_a"):
				if attempt_attack(attack_primary, true):
					get_tree().set_input_as_handled()
					return
			if event.is_action("player_attack_b"):
				if attempt_attack(attack_secondary, false):
					get_tree().set_input_as_handled()
					return
	# Send the input to the current state if we haven't already handled it.
	var new_state = state._input(event)
	# Switch states but don't forward input, because presumably that event was handled.
	if new_state != null:
		state.set_state(new_state)

# Let the current state handle the processing logic; also handle the changing of states.
func _fixed_process(delta):
	if Input.is_action_pressed("player_sprint"):
		if _horizontal_multiplier != walking_multiplier * running_multiplier:
			_horizontal_multiplier = walking_multiplier * running_multiplier
			animation_player.set_speed(_horizontal_multiplier)
	else:
		if _horizontal_multiplier != walking_multiplier:
			_horizontal_multiplier = walking_multiplier
			animation_player.set_speed(_horizontal_multiplier)
	_update_moving_platforms()
	# Update substates.
	for substate in substates:
		substate._fixed_process(delta)
	# Update the current state; handle switching.
	state.state_loop("_fixed_process", delta)

# Graphical processing and state loop.
func _process(delta):
	for substate in substates:
		substate._process(delta)
	state.state_loop("_process", delta)

#################################
# State-based helper functions. #
#################################

# Return true iff the player has feet on a solid.
func is_grounded():
	if _floor_checker.is_overlapping() or _slope_checker_L.is_overlapping() or _slope_checker_R.is_overlapping():
		return true
	return false

# Return true iff there's something immediately above the player at standing height.
func is_roof_blocked():
	return _ceiling_checker.is_overlapping()

# Return true iff a player's slope checker but not their floor checker is on a solid.
func is_on_a_slope():
	if _floor_checker.is_overlapping():
		return false
	if _slope_checker_L.is_overlapping() or _slope_checker_R.is_overlapping():
		return true
	return false

# Return true iff the WallChecker in the direction the player is facing overlaps something.
func is_face_checking_a_wall():
	if direction == 1 and _wall_checker_R.is_overlapping():
		return true
	if direction == -1 and _wall_checker_L.is_overlapping():
		return true
	return false

# If the player wishes to wall jump, return 1 for right or -1 for left.
# Otherwise return 0 to indicate that they should not wall jump.
func should_wall_jump_to():
	# Can't wall jump if you're not pressing jump.
	if not Input.is_action_pressed("player_jump"):
		return 0
	# Can't wall jump if you could regular jump.
	if is_grounded():
		return 0
	# Only a wall on the left and player is holding left or is out of jumps; jump right.
	if _wall_checker_L.is_overlapping() and not _wall_checker_R.is_overlapping():
		if has_extra_jump == false or Input.is_action_pressed("player_left"):
			return 1
	# Only a wall on the right and player is holding right or is out of jumps; jump left.
	if _wall_checker_R.is_overlapping() and not _wall_checker_L.is_overlapping():
		if has_extra_jump == false or Input.is_action_pressed("player_right"):
			return -1
	# No walls or both walls means no wall-jump.
	return 0

# Return the list of things the player could potentially grab.
func could_grab():
	return _grabsphere.overlapping_bodies

# Return true iff the forward checker does not overlap solids but the rear checker does.
func is_walking_downhill():
	var left_grounded = _slope_checker_L.is_overlapping()
	var right_grounded = _slope_checker_R.is_overlapping()
	# Standing on level ground.
	if left_grounded and right_grounded:
		return false
	# Falling.
	if not left_grounded and not right_grounded:
		return false
	# At this point they have opposite values.
	if left_grounded and direction == 1:
		return true
	if right_grounded and direction == -1:
		return true
	return false

# Return true iff there's a small-enough airgap to a solid below.
func should_stick_to_ground():
	var space_state = get_world_2d().get_direct_space_state()
	var global_position = get_global_pos()
	var collisions = space_state.intersect_ray(global_position, Vector2(global_position.x, global_position.y + _vertical_drop_threshold + VERTICAL_GRACE_FACTOR), _raycast_exceptions)
	if collisions.empty():
		return false
	return true

# Return true during states which allow for sticking to moving platforms.
func should_stick_to_platform():
	var state_name = state.get_name()
	if state_name != "Ducking" and state_name != "Standing" and state_name != "Walking":
		if state_name == "Dodging" and is_grounded():
			return true
		else:
			return false
	else:
		return true

# Return true iff the player's hitbox is in a water area.
func is_in_water():
	for area in _hitbox.overlapping_bodies:
		if area.is_in_group("Water"):
			return true
	return false

# Return true iff the player is in a state which allows for attacking.
func can_attack():
	if is_attacking() and _could_alternate_attack == false:
		return false
	var state_name = state.get_name()
	if state_name == "Standing" or state_name == "Walking" or state_name == "Jumping" or state_name == "Falling" or state_name == "Ducking" or state_name == "WallJumping" or state_name == "Dodging":
		return true
	return false

#################################
# Input-based helper functions. #
#################################

# Determine the X component of movement based on the player's input.
func get_horizontal_movement():
	var x_velocity = 0
	var left_horizontal_joystick = Input.get_joy_axis(0, JOY_AXIS_0)
	if abs(left_horizontal_joystick) >= JOYSTICK_THRESHOLD:
		x_velocity = left_horizontal_joystick
	elif Input.is_action_pressed("ui_right"):
		x_velocity = 1
	elif Input.is_action_pressed("ui_left"):
		x_velocity = -1
	return x_velocity * get_horizontal_movement_speed()

# Return how fast the player should move horizontally.
func get_horizontal_movement_speed():
	return horizontal_speed * _horizontal_multiplier

# Given an input event, return a Jump state if the player attempted to jump.
func should_jump(event):
	if (event.type == InputEvent.KEY or event.type == InputEvent.JOYSTICK_BUTTON) and event.is_pressed() and not event.is_echo() and event.is_action("player_jump"):
		return true
	return false

#######################
# Movement Functions. #
#######################

# Move the given player by the given vector, sliding as much as necessary.
func move_by(vector, set_direction=true):
	var initial_position = get_global_pos()
	if set_direction:
		if vector.x > 0:
			set_direction(1)
		elif vector.x < 0:
			set_direction(-1)
	# Stick to downward slopes if we meant to move levelly.
	if vector.y == 0 and is_walking_downhill() and should_stick_to_ground():
		vector.y = abs(vector.x)
	var motion = move(vector)
	if is_colliding():
		# Don't slide if they're grounded and up against a wall.
		if not is_grounded() or not is_face_checking_a_wall() or vector.y < 0:
			var normal = get_collision_normal()
			motion = normal.slide(motion)
			move(motion)
	last_movement_vector = get_global_pos() - initial_position

# Change the player's direction and sprite facing.
func set_direction(new_direction):
	if is_attacking():
		return
	if sign(new_direction) != sign(_sprite.get_scale().x):
		flip_sprite()
	direction = new_direction

# Create/notify the substate for any new MovingPlatforms we stand on.
func _update_moving_platforms():
	if not should_stick_to_platform():
		return
	var current_platform = null
	var current_platform_substate = null
	for substate in substates:
		if substate extends PlatformMoving:
			current_platform = substate.platform
			current_platform_substate = substate
	var arbitrary_new_platform = null
	for body in _floor_checker.overlapping_bodies:
		if body.is_in_group("MovingPlatform"):
			# If we're still overlapping the current platform, don't find a new one.
			if body == current_platform:
				arbitrary_new_platform = null
				break
			arbitrary_new_platform = body
	# If we found a new platform...
	if arbitrary_new_platform != null:
		# transition to it gracefully.
		if current_platform != null:
			current_platform_substate.platform = arbitrary_new_platform
		# Or attach to it for the first time.
		else:
			var substate = PlatformMoving.new()
			substate.platform = arbitrary_new_platform
			substate.attach(self)

#####################
# Attack Functions. #
#####################

# Return true iff the player is attacking.
func is_attacking():
	return attack_animation_player.is_playing()

# Function for animations to signal that an alternate attack would now be allowed.
func can_alternate():
	_could_alternate_attack = true

# Given an attack animation and a handedness, return true iff the appropriate animation player could have that animation started.
func attempt_attack(animation, use_right_hand):
	# Maybe flip animation arms and hands.
	var flip_animation = false
	var front_hand = true
	if (use_right_hand and direction == -1) or (not use_right_hand and direction == 1):
		front_hand = false
		if not _attack_is_flipped[animation]:
			flip_animation = true
	else:
		if _attack_is_flipped[animation]:
			flip_animation = true
	var can_alternate_attack = false
	if _could_alternate_attack and ((front_hand and not _last_attack_front_hand) or (not front_hand and _last_attack_front_hand)):
		can_alternate_attack = true
	# Don't restart animations on the target hand.
	if not attack_animation_player.is_playing() or can_alternate_attack:
		# Stop the animation and flip the arms if we're alternating.
		if flip_animation:
			_flip_attack_animation(animation)
		# Create a weapon if we need one.
		_destroy_weapons()
		_last_attack_front_hand = front_hand
		_could_alternate_attack = false
		if state.get_name() == "Dodging" and is_grounded():
			state.set_state(Standing.new())
		attack_animation_player.play(animation)
		return true
	return false

# Stop whatever the currently running attack is.
func cancel_attack():
	attack_animation_player.stop()
	_destroy_weapons()

# Replace [Left/Right] with its counterpart for all tracks in the given animation.
func _flip_attack_animation(animation_name):
	var animation = attack_animation_player.get_animation(animation_name)
	for i in range(animation.get_track_count()):
		var path = str(animation.track_get_path(i))
		var modified = false
		if path.match("*Left*"):
			path = path.replace("Left", "Right")
			modified = true
		elif path.match("*Right*"):
			path = path.replace("Right", "Left")
			modified = true
		if modified:
			animation.track_set_path(i, NodePath(path))
	_attack_is_flipped[animation_name] = !_attack_is_flipped[animation_name]

# Get the scale that weapons should be multiplied by.
func get_weapon_scale():
	return 1 + (((mind - 100) / 100.0) * 7)

# Callback for animations to have the weapon of their name spawn in the active hand.
func spawn_weapon():
	# Spawn an object based on the animation name, assuming a strict structure.
	var animation_name = attack_animation_player.get_current_animation()
	weapon = load("res://Actors/Player/Weapons/" + animation_name + "/" + animation_name + ".tscn").instance()
	weapon.set_scale(weapon.get_scale() * get_weapon_scale())
	weapon.set_name("Weapon")
	# Apply the weapon to the appropriate hand, flipping its behind_parent logic for the back-hand.
	if _last_attack_front_hand:
		get_node("Sprite/Torso/RightShoulder/RightArm").add_child(weapon)
	else:
		weapon.set_draw_behind_parent(!weapon.is_draw_behind_parent_enabled())
		get_node("Sprite/Torso/LeftShoulder/LeftArm").add_child(weapon)

# Invoke the given method on the player's weapon.
func invoke(function_name):
	assert(weapon != null)
	weapon.call(function_name)

# Destroy the weapon.
func _attack_animation_finished():
	_destroy_weapons()

# Clean up lingering weapons.
func _destroy_weapons():
	# May have fallout from using free().
	if weapon != null:
		weapon.get_node("..").remove_child(weapon)
		weapon.free()
		weapon = null

###########################
# Damage-Taking Functions #
###########################

# TODO
func get_resistances():
	return 0

# TODO
func get_weaknesses():
	return 0

# Take and display damage.
func take_damage(damage_dealer):
	var damage = damage_dealer.get_damage()
	# 50% increase or decrease per matched type.
	var dealt_types = damage_dealer.get_types()
	var net_effectiveness = 0
	for type in DamageDealer.TYPES:
		if type & get_resistances() and type & dealt_types:
			damage = damage / DamageDealer.EFFECTIVENESS
			net_effectiveness = net_effectiveness - 1
		if type & get_weaknesses() and type & dealt_types:
			damage = damage * DamageDealer.EFFECTIVENESS
			net_effectiveness = net_effectiveness + 1
	damage = floor(damage)
	# hp = hp - damage
	# Visually represent the damage.
	var poptext = PopText.instance()
	poptext.set_scale(Vector2(1.5, 1.5))
	poptext.get_node("AnimationPlayer").set_speed(0.5)
	var label = poptext.get_node("Label")
	label.set_text(str(damage))
	if net_effectiveness < 0:
		label.set("custom_colors/font_color", Colors.PLAYER_DAMAGE_RESIST)
	elif net_effectiveness > 0:
		label.set("custom_colors/font_color", Colors.PLAYER_DAMAGE_WEAK)
	else:
		label.set("custom_colors/font_color", Colors.PLAYER_DAMAGE_NORMAL)
	poptext.set_global_pos(Trigonometry.get_global_point_of_overlap(damage_dealer, _hitbox))
	Global.get_level().add_child(poptext)

func _hitbox_area_enter(area):
	if area.is_in_group("DamagesPlayer"):
		state.hit_enemy(area)

#############################
# Shape-Changing Functions. #
#############################

# Multiply the sprite's scale.x by -1.
func flip_sprite():
	var scale = _sprite.get_scale()
	_sprite.set_scale(Vector2(-1 * scale.x, scale.y))

# Return the player to the default shapes.
func set_shape_to_default():
	var collision_shape = get_node("CollisionShape2D")
	collision_shape.get_shape().set_extents(_collision_shape_extents)
	collision_shape.set_pos(Vector2(0, 0))
	get_node("Hitbox/CollisionShape2D").get_shape().set_extents(_hitbox_extents)
	get_node("Hitbox").set_pos(Vector2(0, 0))

# Shrink and lower the collision capsule.
func set_shape_to_ducking():
	var collision_shape = get_node("CollisionShape2D")
	var new_collision_shape_height = _collision_shape_extents.y / DUCKING_DIVISOR
	collision_shape.get_shape().set_extents(Vector2(_collision_shape_extents.x, new_collision_shape_height))
	collision_shape.set_pos(Vector2(0, _collision_shape_extents.y - new_collision_shape_height))
	get_node("Hitbox/CollisionShape2D").get_shape().set_extents(Vector2(_hitbox_extents.x, _hitbox_extents.y / DUCKING_DIVISOR))
	get_node("Hitbox").set_pos(collision_shape.get_pos())

############################
# Stat-oriented functions. #
############################

# Change speeds based on agility.
func _set_agi(new_agi):
	agility = new_agi
	var speed_mod = agility / 100.0
	animation_player.set_speed(speed_mod)
	attack_animation_player.set_speed(speed_mod)
	horizontal_speed = DEFAULT_HORIZONTAL_SPEED * speed_mod
	jump_height = DEFAULT_JUMP_HEIGHT * speed_mod

###############################
# Collision Helper Functions. #
###############################

# Set up collision ignorance for the given object (in both direction).
func ignore_collisions_with(object):
	object.add_collision_exception_with(self)
	_hitbox.ignored_bodies.append(object)
	_grabsphere.ignored_bodies.append(object)
	_floor_checker.ignored_bodies.append(object)
	_slope_checker_L.ignored_bodies.append(object)
	_slope_checker_R.ignored_bodies.append(object)
	_ceiling_checker.ignored_bodies.append(object)
	_wall_checker_L.ignored_bodies.append(object)
	_wall_checker_R.ignored_bodies.append(object)
