extends "res://Actors/Player/States/PlayerState.gd"

const Backflipping = preload("res://Actors/Player/States/Backflipping.gd")
const Ducking = preload("res://Actors/Player/States/Ducking.gd")
const Floating = preload("res://Actors/Player/States/Floating.gd")
const Jumping = preload("res://Actors/Player/States/Jumping.gd")
const SlideKicking = preload("res://Actors/Player/States/SlideKicking.gd")
const Standing = preload("res://Actors/Player/States/Standing.gd")
const Falling = preload("res://Actors/Player/States/Falling.gd")

# 125 / .25 is slightly faster than running.
const DISTANCE = 125 # pixels
const DURATION = 0.25 # s
const HALF_DURATION = DURATION / 2

const UP = Vector2(0, 1) # One-way collision detection is backward.
const DOWN = Vector2(0, -1)

var dodge_backward = false

var _initial_direction = null
var _initial_x_scale = null
var _final_x_scale = null
var _motion = null
var _sprite = null
var _timer = 0

func get_name():
	return "Dodging"

func enter(player):
	.enter(player)
	_sprite = player.get_node("Sprite")
	player.cancel_attack()
	# No dodges immediately after dodging.
	player.has_dodge = false
	# Get the direction from either the keyboard or joystick upfront.
	var joystick_vector = Vector2(Input.get_joy_axis(0, JOY_AXIS_0), Input.get_joy_axis(0, JOY_AXIS_1))
	if joystick_vector.length() >= player.JOYSTICK_THRESHOLD:
		_motion = joystick_vector
	var x_component = 0
	var y_component = 0
	if Input.is_action_pressed("player_up"):
		y_component = -1
	elif Input.is_action_pressed("player_duck"):
		y_component = 1
	if Input.is_action_pressed("player_left"):
		x_component = -1
	elif Input.is_action_pressed("player_right"):
		x_component = 1
	if _motion == null:
		_motion = Vector2(x_component, y_component)
	if dodge_backward:
		_motion = Vector2(-1 * player.direction, 0)
	_motion = _motion.normalized()
	_motion = _motion * DISTANCE
	# Track the scale for procedural animation purposes.
	_initial_direction = player.direction
	_initial_x_scale = _sprite.get_scale().x
	_final_x_scale = -1 * _initial_x_scale

# Allow jump-canceling.
func _input(event):
	if player.should_jump(event) and player.is_grounded():
		player.get_tree().set_input_as_handled()
		# Backflip.
		if Input.is_action_pressed("player_up"):
			return Backflipping.new()
		# Drop through a platform if it had one-way collisions and we're not holding a direction.
		if not Input.is_action_pressed("player_left") and not Input.is_action_pressed("player_right"):
			for body in player.get_node("FloorChecker").overlapping_bodies:
				if body extends PhysicsBody2D:
					if body.get_one_way_collision_direction() == UP:
						player.temporarily_ignore_collisions.append(body)
						body.set_one_way_collision_direction(DOWN)
						return Falling.new()
		# Slide kick if we're not blocked.
		if Input.is_action_pressed("player_duck"):
			if not player.is_roof_blocked() and not player.is_face_checking_a_wall():
				return SlideKicking.new()
			if player.is_roof_blocked():
				return SlideKicking.new()
		# Regular Jump.
		return Jumping.new()

# Move the desired distance.
func _fixed_process(delta):
	# Don't overshoot on the last tick.
	if _timer + delta > DURATION:
		delta = DURATION - _timer
	var percent = delta / DURATION
	player.move_by(_motion * percent)
	_timer = _timer + delta
	if _timer >= DURATION:
		if player.is_in_water():
			return Floating.new()
		if player.is_grounded():
			if Input.is_action_pressed("player_duck"):
				return Ducking.new()
			return Standing.new()
		return Falling.new()

# Animate.
func _process(delta):
	if _timer <= HALF_DURATION:
		_sprite.set_scale(Vector2(lerp(_final_x_scale, _initial_x_scale, 1 - (_timer / HALF_DURATION)), _sprite.get_scale().y))
	else:
		_sprite.set_scale(Vector2(lerp(_initial_x_scale, _final_x_scale, 1 - ((_timer - HALF_DURATION) / HALF_DURATION)), _sprite.get_scale().y))

# Don't get hit while dodging.
func hit_enemy(enemy):
	pass

# Make sure to reset the scale.
func exit():
	_sprite.set_scale(Vector2(_initial_x_scale, _sprite.get_scale().y))
	player.set_direction(_initial_direction)
