extends "res://Actors/Player/States/PlayerState.gd"

const Falling = preload("res://Actors/Player/States/Falling.gd")
const Jumping = preload("res://Actors/Player/States/Jumping.gd")

const INERTIA_MOD = 2
const MAX_FLOAT_SPEED = 100.0 # Pixels / sec.
const X_THRESHOLD = 10

var _hitbox_half_height = null
var _inertia_vector = null
var _player_hitbox = null

func get_name():
	return "Floating"

func enter(player):
	.enter(player)
	# Without delta we can't be precise, so just assume 60FPS upon entry.
	_inertia_vector = player.last_movement_vector * 60
	# Minimum splashdown speed.
	if _inertia_vector.y < MAX_FLOAT_SPEED:
		_inertia_vector.y = MAX_FLOAT_SPEED
	_player_hitbox = player.get_node("Hitbox")
	_hitbox_half_height = _player_hitbox.get_shape(0).get_extents().y / 2.0

# Jump if we're at the top of the waterline.
# Assumes all waterlines are AABBs.
func _input(event):
	if player.should_jump(event):
		if player.get_global_pos().y - _hitbox_half_height <= _get_water_level():
			player.get_tree().set_input_as_handled()
			return Jumping.new()

# Float in a reverse-parabolic fashion until the waterline is reached.
# Allow for horizontal movement.
func _fixed_process(delta):
	# Go sideways out of the water I guess.
	if not player.is_in_water():
		return Falling.new()
	var water_level = _get_water_level()
	var player_pos = player.get_global_pos()
	# Have inertia decay.
	var movement_vector = _inertia_vector
	# x always decays.
	var x = lerp(_inertia_vector.x, 0, INERTIA_MOD * delta)
	if abs(x) < X_THRESHOLD:
		x = 0
	# y only decays if they're submerged.
	var y = _inertia_vector.y
	if player_pos.y > water_level:
		y = lerp(y, -MAX_FLOAT_SPEED, INERTIA_MOD * delta)
	elif y < 0:
		y = 0
	_inertia_vector = Vector2(x, y)
	# Add horizontal component.
	var player_horizontal_movement = player.get_horizontal_movement()
	if player_horizontal_movement != 0:
		movement_vector.x = movement_vector.x + player_horizontal_movement
	# No speed hacks, sadly.
	if abs(movement_vector.x) > player_horizontal_movement:
		movement_vector.x = player_horizontal_movement
	# Then move the player.
	movement_vector = movement_vector * delta
	# (Don't move them above water level.)
	if movement_vector.y < 0 and player_pos.y + movement_vector.y < water_level:
		movement_vector.y = water_level - player_pos.y
	player.move_by(movement_vector)

# Assumes all water objects have an AABB rectangle shape 0.
func _get_water_level():
	var highest_level = null
	for area in _player_hitbox.overlapping_bodies:
		if area.is_in_group("Water"):
			var shape = area.get_shape(0)
			assert(shape extends RectangleShape2D)
			var top_edge = area.get_global_pos().y - (shape.get_extents().y / 2.0) - _hitbox_half_height
			if highest_level == null or top_edge > highest_level:
				highest_level = top_edge
	return highest_level
