extends "res://Actors/Player/States/PlayerState.gd"

const Backflipping = preload("res://Actors/Player/States/Backflipping.gd")
const Grappler = preload("res://Actors/Player/Weapons/Grappler/Grappler.gd")
const Standing = preload("res://Actors/Player/States/Standing.gd")

const BOUNCINESS = 0.75
const MAX_VELOCITY = 5
const REEL_RATE = 480 # pixels / sec
const ROTATION_RATE = PI # rads / sec

const ANGULAR_THRESHOLD = PI / 720.0
const PI_OVER_TWO = PI / 2.0
const TWO_PI = 2 * PI
const THREE_PI_OVER_TWO = 3 * PI / 2.0

# Rope-related.
var _anchor_points = [] # Vector2s
var _level_canvas_rid = null

# Movement-minded.
var _radius = null # From weapon center to anchor point.
var _velocity = null

func get_name():
	return "Grappling"

func enter(player):
	.enter(player)
	_level_canvas_rid = player.get_node("..").get_canvas_item()
	_velocity = player.last_movement_vector.length() # FIXME? Linear, not rotational.
	_radius = (player.weapon.get_global_pos() - _anchor_points[0]).length()
	# Destroy the grappler that got us here.
	player.attack_animation_player.stop(false)

func _input(event):
	# Allow for other attacks.
	if (event.type == InputEvent.KEY or event.type == InputEvent.JOYSTICK_BUTTON) and event.is_pressed() and not event.is_echo():
		if event.is_action("player_attack_a"):
			if player.attempt_attack(player.attack_primary, true):
				player.get_tree().set_input_as_handled()
				_set_state()
		if event.is_action("player_attack_b"):
			if player.attempt_attack(player.attack_secondary, false):
				player.get_tree().set_input_as_handled()
				_set_state()

func _fixed_process(delta):
	# TODO: Gravity, swinging, all that fun stuff.
	if not player.is_grounded():
		_apply_gravity(delta)
	# Reel in/out.
	if Input.is_action_pressed("player_up"):
		_radius = _radius - (REEL_RATE * delta)
		if _radius < 0:
			_radius = 0
	elif Input.is_action_pressed("player_duck"):
		_radius = _radius + (REEL_RATE * delta)
		# TODO: Max Radius
	if Input.is_action_pressed("player_left"):
		_velocity = _velocity + (ROTATION_RATE * delta)
	elif Input.is_action_pressed("player_right"):
		_velocity = _velocity - (ROTATION_RATE * delta)
	_constrain_velocity(delta)
	_move(delta)

func _redraw():
	VisualServer.canvas_item_clear(_level_canvas_rid)
	var last_anchor_point = player.weapon.get_global_pos()
	for anchor_point in _anchor_points:
		# FIXME: Static final imports.
		VisualServer.canvas_item_add_line(_level_canvas_rid, last_anchor_point, anchor_point, Color(1, 0, 0), 3.0)

# Clear the rope drawings when we're done.
func exit():
	VisualServer.canvas_item_clear(_level_canvas_rid)

# Return the angle normalized to 0 <= angle < TWO_PI
func _normalize_angle(angle):
	while angle < 0:
		angle = angle + TWO_PI
	while angle >= TWO_PI:
		angle = angle - TWO_PI
	return angle

# Have the player swing toward a downward face.
func _apply_gravity(delta):
	# No gravity when close enough to downward.
	var angle = _normalize_angle(_angle_to_anchor())
	if angle > (THREE_PI_OVER_TWO - ANGULAR_THRESHOLD) and angle < (THREE_PI_OVER_TWO + ANGULAR_THRESHOLD):
		return
	# Quadrant-based gravity.
	# TODO: Affected by angle.
	if angle >= PI_OVER_TWO and angle < THREE_PI_OVER_TWO:
		_velocity = _velocity + (2 * delta)
	else:
		_velocity = _velocity - (2 * delta)

# Set the player to the appropriate state.
func _set_state():
	# TODO: Other states.
	set_state(Standing.new())

# Add the given point to the list of anchor nodes.
func add_anchor_point(point):
	_anchor_points.push_front(point)

# Return the angle of the player relative to the anchor point.
func _angle_to_anchor():
	var vector = player.weapon.get_global_pos() - _anchor_points[0]
	return atan2(-vector.y, vector.x)

# Minor deceleration and _velocity cap implementation.
func _constrain_velocity(delta):
	# Decelerate.
	if not Input.is_action_pressed("player_left") and not Input.is_action_pressed("player_right"):
		if abs(_velocity) < delta:
			_velocity = 0
		elif _velocity > 0:
			_velocity = _velocity - delta
		else:
			_velocity = _velocity + delta
	# Obey MAX_VELOCITY
	if abs(_velocity) > MAX_VELOCITY:
		_velocity = sign(_velocity) * MAX_VELOCITY

# Move the player toward the target position as best as we're able.
func _move(delta):
	# Don't move if we're already where we want to be.
	var current_position = player.weapon.get_global_pos()
	if _velocity == 0 and abs((current_position - _anchor_points[0]).length() - _radius) <= 1.0:
		return
	# Rotate angles based on _velocity and _radius.
	var desired_angle = _angle_to_anchor() + (_velocity * delta)
	#printt(_radius, _velocity, desired_angle)
	var target_position = _anchor_points[0] + Vector2(_radius * cos(desired_angle), -_radius * sin(desired_angle))
	target_position = target_position - (current_position - player.get_global_pos())
	var motion = player.move_to(target_position)
	if player.is_colliding():
		var normal = player.get_collision_normal()
		print(normal)
		# Either bounce the player or slide them along the wall.
		# FIXME: This logic won't work in most test cases.
		if sign(normal.x) != sign(_velocity):
			_velocity = -BOUNCINESS * _velocity
		else:
			player.move(normal.slide(motion))
			_velocity = 0
	# Update last_movement_vector directly since we moved manually.
	player.last_movement_vector = player.weapon.get_global_pos() - current_position
	_redraw()
