extends "res://Actors/Substate.gd"

# Record all of the player's inputs into a list, then when an action button
# is pressed, scan through them in reverse order (only going back so far)
# and find the most complex move acceptable. Then execute it.

const DragonPunching = preload("res://Actors/Player/States/DragonPunching.gd")
const KamikazeDiving = preload("res://Actors/Player/States/KamikazeDiving.gd")

const INPUT_TIME = 500 # ms

# Simplified action constants.
const UP = 0
const DOWN = 1
const FORWARD = 2
const BACKWARD = 3
const ATTACK = 4

# Special moves (in backward order).
const KAMIKAZE_DIVE = [ATTACK, BACKWARD, BACKWARD, DOWN, FORWARD] # Backward after Backward equals Forward...
const DRAGON_PUNCH = [ATTACK, FORWARD, DOWN, FORWARD]
const FIREBALL = [ATTACK, FORWARD, DOWN]

var _buffer = []

func get_name():
	"Fighting"

func _input(event):
	# Ignore inputs if the player couldn't attack.
	if not parent.can_attack():
		_buffer.clear()
		return
	if (event.type == InputEvent.KEY or event.type == InputEvent.JOYSTICK_BUTTON) and event.is_pressed() and not event.is_echo():
		if event.is_action("player_up"):
			_append_event(UP)
		elif event.is_action("player_duck"):
			_append_event(DOWN)
		elif event.is_action("player_left"):
			if parent.direction == -1:
				_append_event(FORWARD)
			else:
				_append_event(BACKWARD)
		elif event.is_action("player_right"):
			if parent.direction == 1:
				_append_event(FORWARD)
			else:
				_append_event(BACKWARD)
		elif event.is_action("player_attack_a") or event.is_action("player_attack_b"):
			_append_event(ATTACK)
			_consume_events()

# Return true iff the given buffer has the given commands in order, skipping anything in the buffer that doesn't match.
func _contains_commands(buffer, command):
	var buffer_size = buffer.size()
	var command_size = command.size()
	var buffer_index = 0
	var command_index = 0
	while buffer_index < buffer_size:
		if command[command_index] == buffer[buffer_index]:
			command_index = command_index + 1
			if command_index == command_size:
				return true
		buffer_index = buffer_index + 1

# Find the most complex input which ends with the current action, then clear the buffer.
func _consume_events():
	# http://gamedevwithoutacause.com/?p=266
	# Reverse and trim the buffer.
	var backward_buffer = []
	var cutoff_time = OS.get_ticks_msec() - INPUT_TIME
	var i = _buffer.size() - 1
	while i >= 0:
		var event = _buffer[i]
		if event.time < cutoff_time:
			break
		backward_buffer.append(event.event)
		i = i - 1
	_buffer.clear()
	# Match against the various fighting move options, from most to least complex.
	if _contains_commands(backward_buffer, KAMIKAZE_DIVE):
		if _attempt_attack("KamikazeDive") == true:
			parent.state.set_state(KamikazeDiving.new())
	elif _contains_commands(backward_buffer, DRAGON_PUNCH) and parent.is_grounded():
		if _attempt_attack("DragonPunch") == true:
			parent.state.set_state(DragonPunching.new())
	elif _contains_commands(backward_buffer, FIREBALL):
		_attempt_attack("Fireball")
	else:
		print("Punch")

# Attempt an attack with whichever hand is free. Return true iff successful.
func _attempt_attack(attack):
	if parent.attempt_attack(attack, true) == false:
		if parent.attempt_attack(attack, false) == false:
			return false
	return true

# Add the given event at the current time to the buffer.
func _append_event(event):
	var e = Event.new()
	e.event = event
	e.time = OS.get_ticks_msec()
	_buffer.append(e)

# Helper class to remember when events were recorded.
class Event:
	var event = null
	var time = null
