# Class storing the player's knowledge about the game map.
extends Node

const Colors = preload("res://Constants/Colors.gd")
const FileUtil = preload("res://Util/FileUtil.gd")
const Level = preload("res://Levels/Level.gd")
const Zone = preload("res://Levels/Common/Zone.gd")

const _DoorTexture = preload("res://UI/Map/Door.png")
const _FadingCircle = preload("res://UI/Map/FadingCircle.tscn")
const _SquareTexture = preload("res://UI/Map/Square.png")
const _WallTexture = preload("res://UI/Map/Wall.png")

# Directions
const EAST = 1
const NORTH = 2
const WEST = 4
const SOUTH = 8
const DIRECTIONS = [EAST, NORTH, WEST, SOUTH]

# Color semantics.
const REGULAR = Colors.BLUE
const SAVE = Colors.RED

var map_scale = Vector2(1, 1)
var mapped_squares = {} # Vector2 -> Square
var player_coordinates = null

onready var square_size = _SquareTexture.get_size()

onready var _map = Node2D.new()

# Verify no level conflicts during design.
func _ready():
	_DoorTexture.set_flags(0)
	#if OS.is_debug_build():
		#generate_complete_map()

# Register that the player has entered a new zone, possibly mapping it for the first time.
func explore(zone):
	# Establish a player highlight once.
	if player_coordinates == null:
		_map.add_child(_FadingCircle.instance())
	player_coordinates = zone.coordinates
	_map.get_node("FadingCircle").set_pos(player_coordinates * square_size)
	if not mapped_squares.has(player_coordinates):
		var square = _square_from_zone(zone)
		mapped_squares[player_coordinates] = square
		_map_square(player_coordinates, square)

# Return a current representation of the known map.
func get_map():
	return _map

# Regenerate the map representation.
func reset_map():
	if _map != null:
		_map.queue_free()
	_map = Node2D.new()
	for coordinates in mapped_squares.keys():
		_map_square(coordinates, mapped_squares[coordinates])

# Add a new square to our visual representation.
func _map_square(coordinates, square):
	var sprite = Sprite.new()
	sprite.set_centered(false)
	sprite.set_texture(_SquareTexture)
	sprite.set_modulate(square.color)
	for direction in DIRECTIONS:
		if square.walls & direction:
			sprite.add_child(_generate_feature(_WallTexture, direction))
		if square.doors & direction:
			sprite.add_child(_generate_feature(_DoorTexture, direction))
	sprite.set_pos(Vector2(square_size.x * coordinates.x, square_size.y * coordinates.y))
	_map.add_child(sprite)

# Return a sprite with the given texture oriented along the given direction.
func _generate_feature(texture, direction):
	var feature = Sprite.new()
	feature.set_texture(texture)
	if direction != 0:
		feature.set_centered(false)
		if direction == EAST:
			feature.set_pos(Vector2(square_size.x - 1, 0))
		elif direction == NORTH:
			feature.set_pos(Vector2(square_size.x, 0))
			feature.set_rot(-PI / 2.0)
		elif direction == SOUTH:
			feature.set_rot(PI / 2.0)
			feature.set_pos(Vector2(0, square_size.y))
	return feature

# Programatically verify the validity of and generate a complete map based on all levels' zones.
func generate_complete_map():
	mapped_squares.clear()
	for file in FileUtil.find_files_of_type("res://Levels", true, "tscn"):
		var level = load(file).instance()
		if level extends Level:
			# Add it as a child so that its zones are readied.
			# XXX Messes up the camera if done while paused.
			add_child(level)
			_add_all_zones(level)
		level.queue_free()
	reset_map()

# Recursively add all zones at or descended from the current node to the map.
func _add_all_zones(node):
	if node extends Zone:
		mapped_squares[node.coordinates] = _square_from_zone(node)
	for child in node.get_children():
		_add_all_zones(child)

# Validate that the zone describes a new and correct Square, then return it.
func _square_from_zone(zone):
	var coordinates = zone.coordinates
	assert(coordinates != null)
	assert(not mapped_squares.has(coordinates))
	assert(zone.color != null)
	var square = Square.new()
	square.color = zone.color
	square.walls = zone.walls
	square.doors = zone.doors
	return square

# Simple class describing visual representation of a given coordinate.
class Square:
	var color
	var walls
	var doors
