extends Node

const Player = preload("res://Actors/Player/Player.gd")
const Trigonometry = preload("res://Util/Trigonometry.gd")

var _current_scene = null
var _player = null

func _init():
	# Randomize the random seed.
	randomize()

func _ready():
	# Deduce the starting scene.
	var root = get_tree().get_root()
	_current_scene = root.get_child(root.get_child_count() - 1)

# End the current scene, go to another.
func goto_scene(scene):
	_current_scene.queue_free()
	_player = null
	_current_scene = scene.instance()
	get_tree().get_root().add_child(_current_scene)
	get_tree().set_current_scene(_current_scene)
	return _current_scene

# Move the player from the relative coordinates on the source connector to
# the same relative coordinates of the target connector on the target scene.
func transport_player(source_connector, scene, target_connector):
	# Make sure we have a player cached.
	get_player()
	# Track the existing level.
	var level = get_level()
	# Ensure the connectors have the same shape (rectangle) and extents.
	var source_extents = source_connector.get_shape(0).get_extents()
	var is_horizontal = source_connector.is_horizontal
	# Figure out where the player was relative to their exit point.
	var relative_pos = _player.get_global_pos() - source_connector.get_global_pos()
	# Remove the child from its parent
	_player.get_node("..").remove_child(_player)
	# so it doesn't disappear with the level. Which I really hope equals _current_scene.
	_current_scene.queue_free()
	# Now set the _current_scene to the passed-in argument.
	# Its PlayerSpawners should hold off since we still have a _player cached.
	_current_scene = load(scene).instance()
	# Add that scene to the tree, that way all its stuff readies.
	get_tree().get_root().add_child(_current_scene)
	get_tree().set_current_scene(_current_scene)
	# Make sure the _target_connector doesn't spit us right back.
	target_connector = _current_scene.get_node(target_connector)
	#target_connector.disconnect("area_enter", target_connector, "_on_area_enter")
	# Also make sure it had the same shape.
	var extents = target_connector.get_shape(0).get_extents()
	assert(extents == source_extents)
	assert(target_connector.is_horizontal == is_horizontal)
	# XXX: KinematicBody2D? physics goes slightly awry if allowed to fixed_process here.
	_player.set_fixed_process(false)
	# Set the player so that their hitbox is just past the opposite edge from which they entered.
	var player_extents = _player.get_node("Hitbox").get_shape(0).get_extents()
	if is_horizontal == false:
		if relative_pos.x <= 0:
			relative_pos.x = source_extents.x + player_extents.x + 1
		else:
			relative_pos.x = -source_extents.x - player_extents.x - 1
	else:
		if relative_pos.y <= 0:
			relative_pos.y = source_extents.y + player_extents.y + 1
		else:
			relative_pos.y = -source_extents.y - player_extents.y - 1
	_player.set_global_pos(target_connector.get_global_pos() + relative_pos)
	# And add them.
	_current_scene.add_child(_player)
	# Re-enable the physics sometime after the wonkiness ends.
	_player.call_deferred("set_fixed_process", true)

# Reparent the given object to the given new_parent.
# Scale, rotation, and position will be preserved in an intuitive fashion.
# Note that the object's _ready() function will get called.
func reparent(object, new_parent):
	Trigonometry.set_relative_angle(object, new_parent, object.get_rot())
	Trigonometry.set_relative_scale(object, new_parent, object.get_scale())
	var global_pos = object.get_global_pos()
	object.get_node("..").remove_child(object)
	new_parent.add_child(object)
	object.set_global_pos(global_pos)

# Return the first level present as an immediate root of the scene tree (or null).
func get_level():
	for child in get_tree().get_root().get_children():
		if child.is_in_group("Level"):
			return child
	return null

# Find the Player object, cached-BFS-style.
func get_player():
	if _player != null:
		return _player
	var queue = []
	queue.append(get_level())
	while queue.size() > 0:
		var node = queue[0]
		queue.pop_front()
		if node extends Player:
			_player = node
			return node
		for child in node.get_children():
			queue.append(child)
	return null
