const GRAVITY = 98 # Keep current with project settings.

# Set the angle of the object as though it were a direct child of the given ancestor.
static func set_relative_angle(object, ancestor, angle):
	if object == ancestor:
		object.set_rot(angle)
		return
	var reference = object.get_node("..")
	while reference != ancestor:
		# Don't change it if the heirarchy is wrong.
		if reference == null:
			return
		angle = angle - reference.get_rot()
		reference = reference.get_node("..")
	object.set_rot(angle)

# Set the scale of an object as though it were a direct child of the given ancestor.
static func set_relative_scale(object, ancestor, scale):
	if object == ancestor:
		object.set_scale(scale)
		return
	var reference = object.get_node("..")
	while reference != ancestor:
		if reference == null:
			return
		scale = scale * reference.get_scale()
		reference = reference.get_node("..")
	object.set_scale(scale)

# Given a rigid_body_2d, a target_global_position, and an angle, determine the
# velocity vector necessary such that the parabolic projectile arc intercepts
# the target.
static func get_velocity_of_arc_intercept(rigid_body_2d, target_global_position, angle):
	var delta_vector = target_global_position - rigid_body_2d.get_global_pos()
	var gravity = GRAVITY * rigid_body_2d.get_gravity_scale()
	# http://gamedev.stackexchange.com/questions/130619/find-initial-velocity-for-parabolic-arc-with-fixed-angle-that-crosses-target-pos
	var x_velocity = sqrt((gravity * delta_vector.x * delta_vector.x) / (2 * (delta_vector.y - (tan(angle) * delta_vector.x))))
	# Adjust for lost sign.
	if sign(x_velocity) != sign(delta_vector.x):
		x_velocity = -1 * x_velocity
	var y_velocity = x_velocity * tan(angle)
	return Vector2(x_velocity, y_velocity)

# Return the rough point in global coordinates that represents the center
# of the overlap between the AABBs of two objects.
static func get_global_point_of_overlap(target, base):
	var target_aabb = get_aabb(target)
	var base_aabb = get_aabb(base)
	var x_top_left = max(target_aabb.pos.x, base_aabb.pos.x)
	var y_top_left = max(target_aabb.pos.y, base_aabb.pos.y)
	var x_bottom_right = min(target_aabb.end.x, base_aabb.end.x)
	var y_bottom_right = min(target_aabb.end.y, base_aabb.end.y)
	return Vector2(x_top_left + ((x_bottom_right - x_top_left) / 2.0), y_top_left + ((y_bottom_right - y_top_left) / 2.0))

# Given a target object and a base CollisionObject2D, return the point within
# the base object's AABB which is closest to the target's center.
# This point will be relative to the base object.
static func get_bounded_coordinates(target, base):
	# Major oversimplification: use the center of the target object.
	var target_point = target.get_global_pos()
	# Determine the center and extents of the aggregate collision shapes.
	var rectangle = get_aabb(base)
	if OS.is_debug_build():
		Global.get_level().debug_rect = rectangle
	var base_extents = rectangle.size / 2.0
	var base_point = Vector2(rectangle.pos.x + base_extents.x, rectangle.pos.y + base_extents.y)
	# http://math.stackexchange.com/questions/668809/finding-the-coordinates-of-a-point-in-a-given-rectangle
	var x = target_point.x - base_point.x
	var y = target_point.y - base_point.y
	var a = base_extents.x
	var b = base_extents.y
	if x > 0 and abs(a * y) <= abs(b * x):
		return Vector2(a, (a * y) / x)
	elif x < 0 and abs(a * y) <= abs(b * x):
		return Vector2(-a, -((a * y) / x))
	elif y > 0 and abs(a * y) >= abs(b * x):
		return Vector2((b * x) / y, b)
	elif y < 0 and abs(a * y) >= abs(b * x):
		return Vector2(-((b * x) / y), -b)
	assert(false)

# Determine the global-coordinates Rect2D bounding box for the given Area2D.
static func get_aabb(area):
	# Record the furthest x/y extremes we've encountered.
	var max_x = null
	var min_x = null
	var max_y = null
	var min_y = null
	var rid = RID(area)
	var transform = Physics2DServer.area_get_transform(rid)
	var angle = atan2(-transform.x.y, transform.x.x) # nee acos(transform.x.dot(Vector2(1, 0)))
	var scale = Vector2(1, 1)
	var node = area
	while node != null:
		if node.has_method("get_scale"):
			scale = scale * node.get_scale()
		node = node.get_node("..")
	# Determine local bounding-box.
	for i in range(area.get_shape_count()):
		var shape = area.get_shape(i)
		# Don't have to worry about shape scales as those are purely an edit-time convenience.
		if shape extends ConvexPolygonShape2D:
			for point in shape.get_points():
				if max_x == null or point.x > max_x:
					max_x = point.x
				if min_x == null or point.x < min_x:
					min_x = point.x
				if max_y == null or point.y > max_y:
					max_y = point.y
				if min_y == null or point.y < min_y:
					min_y = point.y
		elif shape extends RectangleShape2D:
			var extents = shape.get_extents()
			if max_x == null or extents.x > max_x:
				max_x = extents.x
			if min_x == null or -extents.x < min_x:
				min_x = -extents.x
			if max_y == null or extents.y > max_y:
				max_y = extents.y
			if min_y == null or -extents.y < min_y:
				min_y = -extents.y
		elif shape extends CircleShape2D:
			var radius = shape.get_radius()
			if max_x == null or radius > max_x:
				max_x = radius
			if min_x == null or -radius < min_x:
				min_x = -radius
			if max_y == null or radius > max_y:
				max_y = radius
			if min_y == null or -radius < min_y:
				min_y = -radius
		else:
			print("Shape2D not handled: ", shape)
			assert(false)
	# Globalize that transformation.
	# Rotation.
	var rotated_points = []
	rotated_points.append(_rotate_point(Vector2(min_x, min_y), angle))
	rotated_points.append(_rotate_point(Vector2(min_x, max_y), angle))
	rotated_points.append(_rotate_point(Vector2(max_x, min_y), angle))
	rotated_points.append(_rotate_point(Vector2(max_x, max_y), angle))
	max_x = null
	min_x = null
	max_y = null
	min_y = null
	for point in rotated_points:
		if max_x == null or point.x > max_x:
			max_x = point.x
		if min_x == null or point.x < min_x:
			min_x = point.x
		if max_y == null or point.y > max_y:
			max_y = point.y
		if min_y == null or point.y < min_y:
			min_y = point.y
	# Scale.
	max_x = max_x * scale.x
	min_x = min_x * scale.x
	max_y = max_y * scale.y
	min_y = min_y * scale.y
	# Position.
	max_x = max_x + transform.o.x
	min_x = min_x + transform.o.x
	max_y = max_y + transform.o.y
	min_y = min_y + transform.o.y
	return Rect2(Vector2(min_x, min_y), Vector2(max_x - min_x, max_y - min_y))

# Rotate the given point by the given angle of radians about the origin.
static func _rotate_point(point, angle):
	return Vector2((point.x * cos(angle)) - (point.y * sin(angle)), (point.x * sin(angle)) + (point.y * cos(angle)))
